from cython cimport boundscheck, wraparound, cdivision
from cython.view cimport contiguous
from cython.parallel cimport prange

from cal_tools.enums import BadPixels

ctypedef unsigned short cell_t
ctypedef unsigned short raw_t
ctypedef float data_t
ctypedef unsigned char gain_t
ctypedef unsigned int mask_t

cdef mask_t WRONG_GAIN_VALUE = BadPixels.WRONG_GAIN_VALUE, \
    VALUE_OUT_OF_RANGE = BadPixels.VALUE_OUT_OF_RANGE, \
    VALUE_IS_NAN = BadPixels.VALUE_IS_NAN


@boundscheck(False)
@wraparound(False)
@cdivision(True)
def correct_lpd_frames(
    # (frame, x, y)
    raw_t[:, :, ::contiguous] in_raw,
    cell_t[::contiguous] in_cell,

    # (frame, x, y)
    data_t[:, :, ::contiguous] out_data,
    gain_t[:, :, ::contiguous] out_gain,
    mask_t[:, :, ::contiguous] out_mask,

    # (cell, x, y, gain)
    float[:, :, :, ::contiguous] ccv_offset,
    float[:, :, :, ::contiguous] ccv_gain,
    mask_t[:, :, :, ::contiguous] ccv_mask,

    int num_threads=1,
):
    cdef int frame, ss, fs
    cdef cell_t cell
    cdef data_t data
    cdef gain_t gain
    cdef mask_t mask

    for frame in prange(in_raw.shape[0], nogil=True, num_threads=num_threads):
        cell = in_cell[frame]

        for ss in range(in_raw.shape[1]):
            for fs in range(in_raw.shape[2]):
                # Decode intensity and gain from raw data.
                data = <data_t>(in_raw[frame, ss, fs] & 0xFFF)
                gain = <gain_t>((in_raw[frame, ss, fs] & 0x3000) >> 12)

                if gain <= 2:
                    mask = ccv_mask[cell, ss, fs, gain]
                else:
                    data = 0.0
                    gain = 0
                    mask = WRONG_GAIN_VALUE

                data = data - ccv_offset[cell, ss, fs, gain]
                data = data * ccv_gain[cell, ss, fs, gain]

                if data > 1e7 or data < -1e7:
                    data = 0.0
                    mask = mask | VALUE_OUT_OF_RANGE

                out_data[frame, ss, fs] = data
                out_gain[frame, ss, fs] = gain
                out_mask[frame, ss, fs] = mask
